use std::{error::Error, net::TcpListener, time::Duration};

use arti_client::{config::BoolOrAuto, StreamPrefs, TorClient, TorClientConfig};
use axum::body::Bytes;
use futures::{AsyncReadExt, Stream, StreamExt};
use log::{error, info};
use tor_hsservice::config::OnionServiceConfigBuilder;
use tor_hsservice::handle_rend_requests;
use axum::{
    http::StatusCode,
    Json, Router,
    debug_handler
};
use axum::routing;
use serde::{Deserialize, Serialize};
use futures::io::AsyncWriteExt;
use tor_rtcompat::PreferredRuntime;

use crate::tor_axum;

#[debug_handler]
async fn test_connection_to_tor(payload: Bytes) -> (StatusCode, String)  {
    info!("Got interesting payload! {:#?}", payload);
    (StatusCode::OK, "World is hell indeed!".to_string())
}

#[derive(Serialize, Deserialize)]
struct SimpleMessage {
    msg: String
}

struct OnionHttpClient {
    tor_client: TorClient<PreferredRuntime>
}

impl OnionHttpClient {
    pub async fn request(&self, path: &str, onion_name: String, data: Vec<u8>) -> Result<String, Box<dyn std::error::Error + Send + Sync>> {
        let mut stream_prefs = StreamPrefs::new();
        stream_prefs.connect_to_onion_services(BoolOrAuto::Explicit(true));

        let mut stream  = self.tor_client.connect_with_prefs((onion_name.clone(), 80), &stream_prefs).await?;
        let host = onion_name;

        // Construct the HTTP request manually
        let request = format!(
            "POST {} HTTP/1.1\r\nHost: {}\r\nContent-Type: application/octet-stream\r\nContent-Length: {}\r\nConnection: close\r\n\r\n",
            path, host, data.len()
        ).as_bytes().to_vec();

        let request = [request, data].concat();

        stream.write_all(&request).await?;

        // Flushing the stream is important; see below!
        let _ = stream.flush().await;
        let mut buf = [0; 4096];
        let n = stream.read(&mut buf).await?;
        Ok(String::from_utf8(buf[..n].to_vec()).unwrap())
    }
}

pub async fn new_client() -> Result<OnionHttpClient, Box<dyn Error + Send + Sync>> {
    let tor_client = TorClient::create_bootstrapped(TorClientConfig::default()).await?;
    Ok(OnionHttpClient { tor_client })
}

pub async fn start() -> Result<(), Box<dyn Error>> {
    let tor_client = TorClient::create_bootstrapped(TorClientConfig::default()).await?;

    let (onion_service, rend_requests) = tor_client.launch_onion_service(
        OnionServiceConfigBuilder::default()
            .nickname("gamal-service".to_owned().try_into().unwrap())
            .build()?,
    )?;
    
    let stream_requests = handle_rend_requests(rend_requests);
    
    let app = Router::new().route("/", routing::post(test_connection_to_tor));
    
    println!("serving at: http://{}", onion_service.onion_name().unwrap());
    
    let onion_name = onion_service.onion_name().unwrap();
    tokio::spawn(async move {
        tokio::time::sleep(Duration::from_secs(40)).await;
        if let Ok(client) = new_client().await {
            client.request("/", onion_name.to_string(), [1, 2, 2, 4, 8].to_vec()).await;
        } else {
            error!("Error creating new client!");
        }
    });
    
    tor_axum::serve(stream_requests, app).await;

    Ok(())
}